/*
	svnreauthor version 0.8.0, released 21 Dec 2017

	Copyright 2017 Daniel Suni

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define NEWLINE 10
#define CONTENT_PADDING 2
#define VERSION "0.8.0"

typedef struct {
	char *old_id;
	char *new_id;
	int new_size;
	int diff;
} author;

void exit_with_error(char *message, int exit_code) {
	fprintf(stderr, "ERROR: %s\n", message);
	exit(exit_code);
}

void print_help_and_exit() {
	printf("svnreauthor usage:\n\n");
	printf("svnreauthor oldid_1,newid_1 [oldid_2,newid_2 ...]\n\n");
	printf("Svnreauthor reads svn dump files from stdin and outputs them with changed\n");
	printf("author names to stdout. The old and new author ids are provided as comma\n");
	printf("separated pairs. At least one pair must be provided. Use -v for version.\n\n");
	printf("EXAMPLE:\n\tcat repo.dump | svnreauthor jdoe,johndoe bsmith,williamsmith | svnadmin load newrepo\n\n");
	exit(0);
}

void print_version_and_exit() {
	printf("svnreauthor %s\n", VERSION);
	exit(0);
}

char* str_malloc(size_t sz) {
	char* str;
	if ((str = (char*)malloc(sz)) == NULL) {
		exit_with_error("malloc failed", 2);
	}
	return str;
}

// Returns the number of digits in a number.
int size_of_number(int num) {
	int i = 1;
	while (num >= 10) {
		num /= 10;
		++i;
	}
	return i;
}

// Returns 1 if string a starts with string b, otherwise 0.
int starts_with(char *a, char *b) {
	int i = 0;
	while (b[i] != '\0') {
		if (a[i] != b[i]) {
			return 0;
		}
		++i;
	}
	return 1;
}

void add_author(char *ids, author *auth) {
	int i = 0;
	int oldsz, newsz;
	while (ids[i] != '\0' && ids[i] != ',') {
		++i;
	}
	if (ids[i] == '\0') {
		exit_with_error("Author ids must be comma separated", 1);
	}
	ids[i] = '\0';
	auth->old_id = ids;
	auth->new_id = &ids[i + 1];
	auth->new_size = strlen(auth->new_id);
	if (strlen(auth->old_id) == 0 || auth->new_size == 0) {
		exit_with_error("Ids can not be empty", 1);
	}
	oldsz = size_of_number(strlen(auth->old_id));
	newsz = size_of_number(strlen(auth->new_id));
	auth->diff = (auth->new_size - strlen(auth->old_id)) + (newsz - oldsz);
}

int main(int argc, char **argv) {
	int i, j, nls, c, found;
	off_t prop_len, con_len, i64, j64;
	author *authors = NULL;
	int author_size = argc - 1;
	char *buffer = NULL;
	char *last_line = NULL;
	int buffer_size = 0;
	int line_size = 0;
	int line_len = 0;
	int buffer_len = 0;
	int reading_node = 0;
	int reading_rev = 0;
	int read_buffer = 0;
	int buffer_rows = 0;
	int author_row = 0;
	if (argc < 2) {
		exit_with_error("Insufficient parameters. Use -h for help.", 1);
	}
	if (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0) {
		print_help_and_exit();
	}
	if (strcmp(argv[1], "-v") == 0 || strcmp(argv[1], "--version") == 0) {
		print_version_and_exit();
	}
	for (i = 1 ; i < argc ; ++i) {
		if ((authors = (author*)realloc(authors, i * sizeof(author))) == NULL) {
			exit_with_error("realloc failed", 2);
		}
		add_author(argv[i], &authors[i - 1]);
	}
	while ( (c = getchar()) != EOF ) {
		if (line_size - line_len <= 1) {
			line_size += 100;
			if ((last_line = (char*)realloc(last_line, line_size)) == NULL) {
				exit_with_error("realloc failed", 2);
			}
		}
		last_line[line_len] = c;
		++line_len;
		last_line[line_len] = '\0';
		if (read_buffer) {
			if (buffer_size - buffer_len <= 1) {
				buffer_size += 100;
				if ((buffer = (char*)realloc(buffer, buffer_size)) == NULL) {
					exit_with_error("realloc failed", 2);
				}
			}
			buffer[buffer_len] = c;
			++buffer_len;
			buffer[buffer_len] = '\0';
		}

		if (c == NEWLINE) {
			if (starts_with(last_line, "Revision-number: ") && strcmp(last_line, "Revision-number: 0\n") != 0) {
				read_buffer = 1;
				buffer_rows = 0;
				printf("%s", last_line);
				line_len = 0;
				last_line[0] = '\0';
			}
			else if (read_buffer) {
				++buffer_rows;
				if (starts_with(last_line, "Prop-content-length: ")) {
					prop_len = (off_t)atol(&last_line[21]);
				}
				else if (starts_with(last_line, "Content-length: ")) {
					con_len = (off_t)atol(&last_line[16]);
				}
				else if (strcmp(last_line, "svn:author\n") == 0) {
					author_row = buffer_rows;
				}
			}
			// buffer_rows 2 greater than author_row -> we have all the information we need:
			// (0) svn:author
			// (1) V $length_of_author
			// (2) $author
			if (author_row && buffer_rows - author_row == 2) {
				found = 0;
				last_line[line_len - 1] = '\0'; // We don't want to compare the newline character.
				for (i = 0; i < author_size; ++i) {
					if (strcmp(authors[i].old_id, last_line) == 0) {
						found = 1;
						printf("Prop-content-length: %d\n", prop_len + authors[i].diff);
						printf("Content-length: %d\n", con_len + authors[i].diff);
						j = 0;
						nls = 0;
						while (nls < 2) {
							if (buffer[j] == NEWLINE) {
								++nls;
							}
							++j;
						}
						while (nls < author_row) {
							if (buffer[j] == NEWLINE) {
								++nls;
							}
							fputc(buffer[j], stdout);
							++j;
						}
						printf("V %d\n", authors[i].new_size);
						printf("%s\n", authors[i].new_id);
						break;
					}
				}
				if (!found) {
					printf("%s", buffer);
				}
				read_buffer = 0;
				buffer_rows = 0;
				author_row = 0;
				buffer_len = 0;
				buffer[0] = '\0';
			}
			else if (!read_buffer) {
				printf("%s", last_line);
				if (starts_with(last_line, "Content-length: ")) {
					i64 = (off_t)atol(&last_line[16]) + CONTENT_PADDING;
					j64 = 0;
					while (i64 > j64) {
						fputc(getchar(), stdout);
						++j64;
					}
				}
			}
			line_len = 0;
			last_line[0] = '\0';
		} // if (c == NEWLINE)
	}
	free(buffer);
	free(last_line);
	free(authors);
	return 0;
}
